@@ -20,7 +20,9 @@ from generic_camunda_client import ApiException, CountResultDto, DeploymentWithD
     VariableValueDto, FetchExternalTasksDto, FetchExternalTaskTopicDto, ProcessDefinitionApi, \
     ProcessInstanceWithVariablesDto, StartProcessInstanceDto, ProcessInstanceModificationInstructionDto, \
     ProcessInstanceApi, ProcessInstanceDto, VersionApi, EvaluateDecisionDto, MessageApi, \
-    MessageCorrelationResultWithVariableDto, CorrelationMessageDto
+    MessageCorrelationResultWithVariableDto, CorrelationMessageDto, ActivityInstanceDto, ExternalTaskFailureDto, \
+    IncidentApi, IncidentDto
+
 import generic_camunda_client as openapi_client
 
 # local imports
@@ -68,7 +70,7 @@ class CamundaLibrary:
 
     == Subscribing a topic / long polling ==
     You may achieve a kind of subscription by providing the ``asyncResponseTimeout`` with the `Fetch workload`
-    keyword in order to achieve [https://docs.camunda.org/manual/7.14/user-guide/process-engine/external-tasks/#long-polling-to-fetch-and-lock-external-tasks|Long Polling].
+    keyword in order to achieve [https://docs.camunda.org/manual/latest/user-guide/process-engine/external-tasks/#long-polling-to-fetch-and-lock-external-tasks|Long Polling].
 
     | ${variables} | fetch workload | my_topic | async_response_timeout=60000 |
     | log | Waited at most 1 minute before this log statement got executed |
@@ -81,19 +83,21 @@ class CamundaLibrary:
     = Feedback =
 
     Feedback is very much appreciated regardless if it is comments, reported issues, feature requests or even merge
-    requests. You are welcome to participating in any way at the [https://gitlab.com/postadress/robotframework/robotframework-camunda|GitLab project of CamundaLibrary].
+    requests. You are welcome to participating in any way at the [https://github.com/MarketSquare/robotframework-camunda|GitHub project of CamundaLibrary].
     """
 
     WORKER_ID = f'robotframework-camundalibrary-{time.time()}'
 
     EMPTY_STRING = ""
-    KNOWN_TOPICS: Dict[str,Dict[str, Any]] = {}
+    KNOWN_TOPICS: Dict[str, Dict[str, Any]] = {}
     FETCH_RESPONSE: LockedExternalTaskDto = {}
+    DEFAULT_LOCK_DURATION = None
 
     def __init__(self, camunda_engine_url: str = 'http://localhost:8080'):
         self._shared_resources = CamundaResources()
         if camunda_engine_url:
             self.set_camunda_url(camunda_engine_url)
+        self.DEFAULT_LOCK_DURATION = self.reset_task_lock_duration()
 
     @keyword("Set Camunda URL")
     def set_camunda_url(self, url: str):
@@ -105,6 +109,34 @@ class CamundaLibrary:
             raise ValueError('Cannot set camunda engine url: no url given.')
         self._shared_resources.camunda_url = url_normalize(f'{url}/engine-rest')
 
+    @keyword("Set Task Lock Duration", tags=["task"])
+    def set_task_lock_duration(self, lock_duration: int):
+        """
+        Sets lock duration used as default when fetching. Camunda locks a process instance for the period. If the
+        external task os not completed, Camunda gives the process instance free for another attempt only when lock
+        duration has expired.
+
+        Value is in milliseconds (1000 = 1 minute)
+        """
+        try:
+            self.DEFAULT_LOCK_DURATION = int(lock_duration)
+        except ValueError:
+            logger.error(f'Failed to set lock duration. Value does not seem a valid integer:\t{lock_duration}')
+
+    @keyword("Reset Task Lock Duration", tags=["task"])
+    def reset_task_lock_duration(self):
+        """
+        Counter keyword for "Set Task Lock Duration". Resets lock duration to the default. The default is either
+        environment variable CAMUNDA_TASK_LOCK_DURATION or 600000 (10 minutes).
+        """
+        try:
+            lock_duration = int(os.environ.get('CAMUNDA_TASK_LOCK_DURATION', 600000))
+        except ValueError as e:
+            logger.warn(
+                f'Failed to interpret "CAMUNDA_TASK_LOCK_DURATION". Environment variable does not seem to contain a valid integer:\t{e}')
+            lock_duration = 600000
+        return lock_duration
+
     @keyword("Get Camunda URL")
     def get_camunda_url(self) -> str:
         return self._shared_resources.camunda_url
@@ -318,10 +350,10 @@ class CamundaLibrary:
             # Create an instance of the API class
             api_instance = openapi_client.ExternalTaskApi(api_client)
             if 'lock_duration' not in kwargs:
-                kwargs['lock_duration'] = 60000
+                kwargs['lock_duration'] = self.DEFAULT_LOCK_DURATION
             if 'deserialize_values' not in kwargs:
                 kwargs['deserialize_values'] = False
-            topic_dto=FetchExternalTaskTopicDto(topic_name=topic, **kwargs)
+            topic_dto = FetchExternalTaskTopicDto(topic_name=topic, **kwargs)
             fetch_external_tasks_dto = FetchExternalTasksDto(worker_id=self.WORKER_ID, max_tasks=1,
                                                              async_response_timeout=async_response_timeout,
                                                              use_priority=use_priority,
@@ -388,10 +420,82 @@ class CamundaLibrary:
             return self.FETCH_RESPONSE.to_dict()
         return self.FETCH_RESPONSE
 
-    @keyword("Drop fetch response")
+    @keyword("Drop fetch response", tags=['task'])
     def drop_fetch_response(self):
         self.FETCH_RESPONSE = {}
 
+    @keyword("Throw BPMN Error", tags=['task'])
+    def bpmn_error(self, error_code: str, error_message: str = None, variables: Dict[str, Any] = None,
+                   files: Dict = None):
+        if not self.FETCH_RESPONSE:
+            logger.warn('No task to complete. Maybe you did not fetch and lock a workitem before?')
+        else:
+            with self._shared_resources.api_client as api_client:
+                api_instance = openapi_client.ExternalTaskApi(api_client)
+                variables = CamundaResources.convert_dict_to_openapi_variables(variables)
+                openapi_files = CamundaResources.convert_file_dict_to_openapi_variables(files)
+                variables.update(openapi_files)
+                bpmn_error = openapi_client.ExternalTaskBpmnError(worker_id=self.WORKER_ID,
+                                                                  error_message=error_message,
+                                                                  error_code=error_code,
+                                                                  variables=variables)
+                try:
+                    logger.debug(f"Sending BPMN error for task:\n{bpmn_error}")
+                    api_instance.handle_external_task_bpmn_error(self.FETCH_RESPONSE.id,
+                                                                 external_task_bpmn_error=bpmn_error)
+                    self.drop_fetch_response()
+                except ApiException as e:
+                    logger.error(f"Exception when calling ExternalTaskApi->handle_external_task_bpmn_error: {e}\n")
+
+    @keyword("Notify failure", tags=["task"])
+    def notify_failure(self, **kwargs):
+        """
+        Raises a failure to Camunda. When retry counter is less than 1, an incident is created by Camunda.
+
+        CamundaLibrary takes care of providing the worker_id and task_id. *retry_timeout* is equal to *lock_duration* for external tasks.
+        Check for camunda client documentation for all parameters of the request body: https://noordsestern.gitlab.io/camunda-client-for-python/7-15-0/docs/ExternalTaskApi.html#handle_failure
+
+        Example:
+        | *notify failure* |  |  |
+        | *notify failure* | retries=3 | error_message=Task failed due to... |
+        """
+        if not self.FETCH_RESPONSE:
+            logger.warn('No task to notify failure for. Maybe you did not fetch and lock a workitem before?')
+        else:
+            with self._shared_resources.api_client as api_client:
+                api_instance = openapi_client.ExternalTaskApi(api_client)
+                external_task_failure_dto = ExternalTaskFailureDto(worker_id=self.WORKER_ID, retry_timeout=60000,
+                                                                   **kwargs)
+
+                try:
+                    api_instance.handle_failure(id=self.FETCH_RESPONSE.id,
+                                                external_task_failure_dto=external_task_failure_dto)
+                    self.drop_fetch_response()
+                except ApiException as e:
+                    logger.error("Exception when calling ExternalTaskApi->handle_failure: %s\n" % e)
+
+    @keyword("Get incidents")
+    def get_incidents(self, **kwargs):
+        """
+        Retrieves incidents matching given filter arguments.
+
+        For full parameter list checkout: https://noordsestern.gitlab.io/camunda-client-for-python/7-15-0/docs/IncidentApi.html#get_incidents
+
+        Example:
+        | ${all_incidents} | *get incidents* |  |
+        | ${incidents_of_process_instance | *get incidentse* | process_instance_id=${process_instance}[process_instance_id] |
+        """
+        with self._shared_resources.api_client as api_client:
+            api_instance: IncidentApi = openapi_client.IncidentApi(api_client)
+
+            try:
+                response: List[IncidentDto] = api_instance.get_incidents(**kwargs)
+            except ApiException as e:
+                logger.error(f'Failed to get incidents:\n{e}')
+                raise e
+
+        return [incident.to_dict() for incident in response]
+
     @keyword("Complete task", tags=['task'])
     def complete(self, result_set: Dict[str, Any] = None, files: Dict = None):
         """
@@ -427,10 +531,12 @@ class CamundaLibrary:
                 variables = CamundaResources.convert_dict_to_openapi_variables(result_set)
                 openapi_files = CamundaResources.convert_file_dict_to_openapi_variables(files)
                 variables.update(openapi_files)
-                complete_task_dto = openapi_client.CompleteExternalTaskDto(worker_id=self.WORKER_ID, variables=variables)
+                complete_task_dto = openapi_client.CompleteExternalTaskDto(worker_id=self.WORKER_ID,
+                                                                           variables=variables)
                 try:
                     logger.debug(f"Sending to Camunda for completing Task:\n{complete_task_dto}")
-                    api_instance.complete_external_task_resource(self.FETCH_RESPONSE.id, complete_external_task_dto=complete_task_dto)
+                    api_instance.complete_external_task_resource(self.FETCH_RESPONSE.id,
+                                                                 complete_external_task_dto=complete_task_dto)
                     self.drop_fetch_response()
                 except ApiException as e:
                     logger.error(f"Exception when calling ExternalTaskApi->complete_external_task_resource: {e}\n")
@@ -469,7 +575,7 @@ class CamundaLibrary:
 
     @keyword("Start process", tags=['process'])
     def start_process(self, process_key: str, variables: Dict = None, files: Dict = None,
-                      before_activity_id: str = None, after_activity_id: str = None,**kwargs) -> Dict:
+                      before_activity_id: str = None, after_activity_id: str = None, **kwargs) -> Dict:
         """
         Starts a new process instance from a process definition with given key.
 
@@ -538,20 +644,68 @@ class CamundaLibrary:
                 raise e
 
     @keyword("Get all active process instances", tags=['process'])
-    def get_all_process_instances(self, process_definition_key):
+    def get_all_active_process_instances(self, process_definition_key):
         """
         Returns a list of process instances that are active for a certain process definition identified by key.
         """
+        return self.get_process_instances(process_definition_key=process_definition_key, active='true')
+
+    @keyword("Get process instances")
+    def get_process_instances(self, **kwargs):
+        """
+        Queries Camunda for process instances that match certain criteria.
+
+        Be aware, that boolean value must be strings either 'true' or 'false'
+
+         async_req: execute request asynchronously
+         sort_by: Sort the results lexicographically by a given criterion. Must be used in conjunction with the sortOrder parameter.
+         sort_order: Sort the results in a given order. Values may be asc for ascending order or desc for descending order. Must be used in conjunction with the sortBy parameter.
+         first_result: Pagination of results. Specifies the index of the first result to return.
+         max_results: Pagination of results. Specifies the maximum number of results to return. Will return less results if there are no more results left.
+         process_instance_ids: Filter by a comma-separated list of process instance ids.
+         business_key: Filter by process instance business key.
+         business_key_like: Filter by process instance business key that the parameter is a substring of.
+         case_instance_id: Filter by case instance id.
+         process_definition_id: Filter by the deployment the id belongs to.
+         process_definition_key: Filter by the key of the process definition the instances run on.
+         process_definition_key_in: Filter by a comma-separated list of process definition keys. A process instance must have one of the given process definition keys.
+         process_definition_key_not_in: Exclude instances by a comma-separated list of process definition keys. A process instance must not have one of the given process definition keys.
+         deployment_id: Filter by the deployment the id belongs to.
+         super_process_instance: Restrict query to all process instances that are sub process instances of the given process instance. Takes a process instance id.
+         sub_process_instance: Restrict query to all process instances that have the given process instance as a sub process instance. Takes a process instance id.
+         super_case_instance: Restrict query to all process instances that are sub process instances of the given case instance. Takes a case instance id.
+         sub_case_instance: Restrict query to all process instances that have the given case instance as a sub case instance. Takes a case instance id.
+         active: Only include active process instances. Value may only be true, as false is the default behavior.
+         suspended: Only include suspended process instances. Value may only be true, as false is the default behavior.
+         with_incident: Filter by presence of incidents. Selects only process instances that have an incident.
+         incident_id: Filter by the incident id.
+         incident_type: Filter by the incident type. See the [User Guide](https://docs.camunda.org/manual/7.15/user-guide/process-engine/incidents/#incident-types) for a list of incident types.
+         incident_message: Filter by the incident message. Exact match.
+         incident_message_like: Filter by the incident message that the parameter is a substring of.
+         tenant_id_in: Filter by a comma-separated list of tenant ids. A process instance must have one of the given tenant ids.
+         without_tenant_id: Only include process instances which belong to no tenant.
+         process_definition_without_tenant_id: Only include process instances which process definition has no tenant id.
+         activity_id_in: Filter by a comma-separated list of activity ids. A process instance must currently wait in a leaf activity with one of the given activity ids.
+         root_process_instances: Restrict the query to all process instances that are top level process instances.
+         leaf_process_instances: Restrict the query to all process instances that are leaf instances. (i.e. don't have any sub instances).
+         variables: Only include process instances that have variables with certain values. Variable filtering expressions are comma-separated and are structured as follows:  A valid parameter value has the form `key_operator_value`. `key` is the variable name, `operator` is the comparison operator to be used and `value` the variable value.  **Note**: Values are always treated as String objects on server side.  Valid `operator` values are: `eq` - equal to; `neq` - not equal to; `gt` - greater than; `gteq` - greater than or equal to; `lt` - lower than; `lteq` - lower than or equal to; `like`. `key` and `value` may not contain underscore or comma characters.
+         variable_names_ignore_case: Match all variable names in this query case-insensitively. If set to true variableName and variablename are treated as equal.
+         variable_values_ignore_case: Match all variable values in this query case-insensitively. If set to true variableValue and variablevalue are treated as equal.
+         _preload_content: if False, the urllib3.HTTPResponse object will
+                                 be returned without reading/decoding response
+                                 data. Default is True.
+         _request_timeout: timeout setting for this request. If one
+                                 number provided, it will be total request
+                                 timeout. It can also be a pair (tuple) of
+                                 (connection, read) timeouts.
+        """
         with self._shared_resources.api_client as api_client:
             api_instance: ProcessInstanceApi = openapi_client.ProcessInstanceApi(api_client)
 
             try:
-                response: List[ProcessInstanceDto] = api_instance.get_process_instances(
-                    process_definition_key=process_definition_key,
-                    active='true'
-                )
+                response: List[ProcessInstanceDto] = api_instance.get_process_instances(**kwargs)
             except ApiException as e:
-                logger.error(f'Failed to get process instances of process {process_definition_key}:\n{e}')
+                logger.error(f'Failed to get process instances of process:\n{e}')
                 raise e
 
         return [process_instance.to_dict() for process_instance in response]
@@ -590,7 +744,7 @@ class CamundaLibrary:
         return response
 
     @keyword("Get Activity Instance", tags=['process'])
-    def get_activity_instance(self, **kwargs):
+    def get_activity_instance(self, id: str):
         """
         Returns an Activity Instance (Tree) for a given process instance.
 
@@ -603,10 +757,11 @@ class CamundaLibrary:
             api_instance: ProcessInstanceApi = openapi_client.ProcessInstanceApi(api_client)
 
             try:
-                response = api_instance.get_activity_instance_tree(**kwargs)
+                response: ActivityInstanceDto = api_instance.get_activity_instance_tree(id)
             except ApiException as e:
-                logger.error(f'failed to get activity tree for process instance {kwargs}:\n{e}')
-        return response
+                logger.error(f'failed to get activity tree for process instance with id {id}:\n{e}')
+                raise e
+        return response.to_dict()
 
     @keyword("Get Process Instance Variable", tags=['process'])
     def get_process_instance_variable(self, process_instance_id: str, variable_name: str):
